self.__BUILD_MANIFEST=function(s,c,e,a,t,f,u,n,o,b,j,r,i,k,d){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":["static/chunks/pages/index-444f1804401f04ea.js"],"/_error":["static/chunks/pages/_error-c66a4e8afc46f17b.js"],"/clusters":["static/chunks/pages/clusters-57221ec2e4e01076.js"],"/clusters/[cluster]":[s,c,e,f,u,j,b,a,t,n,r,o,i,k,d,"static/chunks/6856-5c94d394259cdb6e.js","static/chunks/1871-df9f87fcb7f24292.js","static/chunks/pages/clusters/[cluster]-18b334dedbd9f6f2.js"],"/clusters/[cluster]/[job]":[s,c,e,f,a,t,o,"static/chunks/pages/clusters/[cluster]/[job]-602eeead010ec1d6.js"],"/config":["static/chunks/pages/config-dfb9bf07b13045f4.js"],"/infra":["static/chunks/pages/infra-872e6a00165534f4.js"],"/infra/[context]":["static/chunks/pages/infra/[context]-44ce535a0a0ad4ec.js"],"/jobs":["static/chunks/pages/jobs-0dc34cf9a8710a9f.js"],"/jobs/pools/[pool]":[s,c,e,u,b,a,t,n,"static/chunks/pages/jobs/pools/[pool]-e020fd69dbe76cea.js"],"/jobs/[job]":[s,c,e,f,u,b,a,t,n,o,"static/chunks/pages/jobs/[job]-8677af16befde039.js"],"/users":["static/chunks/pages/users-3a543725492fb896.js"],"/volumes":["static/chunks/pages/volumes-d2af9d22e87cc4ba.js"],"/workspace/new":["static/chunks/pages/workspace/new-3f88a1c7e86a3f86.js"],"/workspaces":["static/chunks/pages/workspaces-6fc994fa1ee6c6bf.js"],"/workspaces/[name]":[s,c,e,f,u,j,a,t,n,r,o,i,k,d,"static/chunks/1141-ec6f902ffb865853.js","static/chunks/pages/workspaces/[name]-9ad108cd67d16d96.js"],sortedPages:["/","/_app","/_error","/clusters","/clusters/[cluster]","/clusters/[cluster]/[job]","/config","/infra","/infra/[context]","/jobs","/jobs/pools/[pool]","/jobs/[job]","/users","/volumes","/workspace/new","/workspaces","/workspaces/[name]"]}}("static/chunks/616-3d59f75e2ccf9321.js","static/chunks/6130-2be46d70a38f1e82.js","static/chunks/5739-d67458fcb1386c92.js","static/chunks/6989-01359c57e018caa4.js","static/chunks/3850-ff4a9a69d978632b.js","static/chunks/7411-b15471acd2cba716.js","static/chunks/1272-1ef0bf0237faccdb.js","static/chunks/8969-0389e2cb52412db3.js","static/chunks/6135-4b4d5e824b7f9d3c.js","static/chunks/6212-7bd06f60ba693125.js","static/chunks/7359-c8d04e06886000b3.js","static/chunks/6990-f6818c84ed8f1c86.js","static/chunks/4282-d2f3ef2fbf78e347.js","static/chunks/6601-06114c982db410b6.js","static/chunks/3015-d014dc5b9412fade.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();